extends Node3D

var Disk = preload("res://elements/disk.tscn")

const RAY_LENGTH = 5

@onready var disks = get_node("disks")

@export var disk_count: int = 1

var movement_count = 0

# Stores the cylinder the disks started in
var starting_cylinder

# Stores the origin cylinder from a drag & drop movement
var origin_cylinder

# Set if the input should be disabled for the game (for pause/win scenes)
var input_disabled: bool = false

signal finished_movement

func reset():
	movement_count = 0
	origin_cylinder = null
	
	for disk in disks.get_children():
		disk.queue_free()
	
	for c in get_node("cylinders").get_children():
		c.reset()
		
	_init_disks()

func _get_random_cylinder():
	var cylinders = get_node("cylinders").get_children()
	var origin_cylinder_index = randi_range(0, len(cylinders)-1)
	return cylinders[origin_cylinder_index]

func set_origin_cylinder(cylinder):
	var cylinder_node = get_node("cylinders/" + cylinder)
	if cylinder_node == null:
		cylinder_node = get_node("cylinders/left")
		
	origin_cylinder = cylinder_node
	reset()
	
	return cylinder_node

func get_origin_cylinder():
	if origin_cylinder == null:
		origin_cylinder = _get_random_cylinder()
	return origin_cylinder

func _init_disks():
	var cylinder = get_origin_cylinder()
	
	for n in range(disk_count, 0, -1):
		var disk = Disk.instantiate()
		disk.size = n
		disk.position.x = cylinder.position.x
		disk.position.y = 3
		disks.add_child(disk)
		cylinder.append_disk(disk)

func _ready():
	GlobalEngine.game_paused.connect(_on_game_paused)
	GlobalEngine.game_unpaused.connect(_on_game_unpaused)
	GlobalEngine.game_finished.connect(_on_game_finished)
	GlobalEngine.game_reset.connect(_on_game_reset)
		
	reset()
	
func _physics_process(_delta):
	if Input.is_action_just_pressed("click") or Input.is_action_just_released("click"):
		_calculate_input()

func _calculate_input():
	if input_disabled:
		return
	
	var hovering_cylinder
	var space_state = get_world_3d().direct_space_state
	var cam = get_viewport().get_camera_3d()
	var mousepos = get_viewport().get_mouse_position()

	var origin = cam.project_ray_origin(mousepos)
	var end = origin + cam.project_ray_normal(mousepos) * RAY_LENGTH
	var query = PhysicsRayQueryParameters3D.create(origin, end)
	query.collide_with_areas = true
	var result = space_state.intersect_ray(query)
	
	if result:
		hovering_cylinder = result.collider
			
	if Input.is_action_pressed("click"):
		if hovering_cylinder:
			var disk = hovering_cylinder.get_top_disk()
			if disk:
				disk.set_follow_mouse(true)
				origin_cylinder = hovering_cylinder
				
	if Input.is_action_just_released("click"):
		if not origin_cylinder:
			return
		
		if not hovering_cylinder:
			var disk = origin_cylinder.get_top_disk()
			if disk:
				disk.set_follow_mouse(false)
				disk.tween_to(origin_cylinder.get_position_for_disk(disk))
				origin_cylinder = null
			return
		
		var disk = origin_cylinder.get_top_disk()
		if disk:
			disk.set_follow_mouse(false)
			if hovering_cylinder.can_append_disk(disk):
				hovering_cylinder.append_disk(origin_cylinder.pop_top_disk())
				finished_movement.emit(disk, hovering_cylinder)
			else:
				disk.tween_to(origin_cylinder.get_position_for_disk(disk))

			origin_cylinder = null

func _on_finished_movement(_disk, cylinder):
	movement_count += 1
	
	if len(cylinder.disks) == disk_count and cylinder != starting_cylinder:
		# TODO: elapsed time
		print("Game finished!")
		GlobalEngine.finish_game(0, movement_count)

func _on_game_paused():
	input_disabled = true

func _on_game_unpaused():
	input_disabled = false
	
func _on_game_finished():
	input_disabled = true
	
func _on_game_reset():
	reset()
	#input_disabled = false
